module KoL.Command where

import Prelude hiding (read, catch)

import Logging
import KoL.Util
import KoL.UtilTypes
import KoL.Api

import Data.IORef
import Control.Concurrent
import Control.Exception

import State
import qualified Data.Map

doCOMMAND_DEBUG _ = return ()
-- doCOMMAND_DEBUG x = putStrLn $ "COMMAND DEBUG: " ++ x

get_stid ref = do
	doCOMMAND_DEBUG "get_stid"
	let Just sessid = get_sessid $ cookie $ connection $ ref
	ai <- getApiInfo ref
	return (charName ai, ascension ai, daysthisrun ai, sessid)

loadState ref = do
	doCOMMAND_DEBUG "loadState"
	newstid <- get_stid ref
-- 	log_time_interval ref "printing newstid line" $ putStrLn $ "loading state for... " ++ show newstid
	stv <- readIORef $ stateData $ sessionData $ ref
	let usable_old_st = case stv of
		Just (stid, oldst) -> if stid == newstid then Just oldst else Nothing
		Nothing -> Nothing
	case usable_old_st of
		Just x -> return x
		Nothing -> log_time_interval ref "read state from db" $ do
			putStrLn $ "loading state from DB..."
			mv <- newEmptyMVar
			(doStateAction $ sessionData $ ref) ref $ \db -> do
-- 				let Just sessid = get_sessid $ cookie $ connection $ ref
				val <- try $ do
					ai <- getApiInfo ref
					let readMapFromDB base_tablename = do
						tablename <- get_state_tablename ref base_tablename
						do_db_query db ("CREATE TABLE IF NOT EXISTS " ++ tablename ++ "(name TEXT NOT NULL, value TEXT NOT NULL, PRIMARY KEY(name));") []
						rs <- do_db_query db ("SELECT name, value FROM " ++ tablename ++ ";") []
						let mapper x = case x of
							[Just name, Just value] -> (name, value)
							_ -> throw $ InternalError $ "Error loading state from database"
						return $ Data.Map.fromList $ map mapper rs

					sessionmap <- readMapFromDB "session"
					sessionmap <- if Data.Map.member "character" sessionmap
						then return sessionmap
						else do
							putStrLn $ "DB:Retrieving session data..."
							putStrLn $ "DB:retrieved name=" ++ (charName ai)
							sessionmap <- return $ Data.Map.insert "character" (charName ai) Data.Map.empty
							sessionmap <- return $ Data.Map.insert "pwd" (pwd ai) sessionmap
							sessionmap <- return $ Data.Map.insert "ascension number" (show $ ascension ai) sessionmap
							sessionmap <- return $ Data.Map.insert "day" (show $ daysthisrun ai) sessionmap
							return sessionmap

					requestmap <- return Data.Map.empty
					charmap <- readMapFromDB "character"
					ascmap <- readMapFromDB "ascension"
					daymap <- readMapFromDB "day"
					return (requestmap, sessionmap, charmap, ascmap, daymap)
				putMVar mv val
			maybenewst <- takeMVar mv
			case maybenewst of
				Right newst -> do
					writeIORef (luaInstances $ sessionData $ ref) Data.Map.empty
					writeIORef (stateData $ sessionData $ ref) $ Just (newstid, newst)
					return newst
				Left err -> do
					putStrLn $ "DEBUG loadState exception: " ++ show err
					throwIO (err :: SomeException)
