module State where

-- State should be: Character, Acension, Day. Gamesession, kolproxysession, pageload.

import Prelude hiding (read, catch)

import PlatformLowlevel

import KoL.Util
import KoL.UtilTypes
import KoL.Api
import KoL.Connection

import Data.Maybe
import Data.IORef
import Control.Concurrent
import Control.Exception
import Control.Monad
import Control.Applicative

import Data.List
import qualified Data.Map
import qualified Data.ByteString.Char8

import System.IO.Error (isDoesNotExistError)

import Text.JSON
import Text.Printf

doSTATE_DEBUG = return
-- doSTATE_DEBUG x = putStrLn $ "STATE DEBUG: " ++ x

readMapFromFile filename = do
	doSTATE_DEBUG $ "readMapFromFile: " ++ filename
	catchJust (\e -> if isDoesNotExistError e then Just e else Nothing)
		(do
			path <- getDirectoryPath "state" filename
			foo_raw_data <- readFile path
			foo_raw <- length foo_raw_data `seq` return foo_raw_data -- force file to be read before closing
			let Just foo = read_as foo_raw :: Maybe (Data.Map.Map String String)
			return foo)
		(\_e -> do
			putStrLn $ "State file " ++ filename ++ " does not already exist."
			return Data.Map.empty)

writeMapToFile filename m = do
	doSTATE_DEBUG $ "writeMapToFile: " ++ filename
	let filedata = show m
-- 	log_debug_msg $ "write-map: " ++ (show filename) ++ " [" ++ (show $ length filedata) ++ "]"
	path <- getDirectoryPath "state" filename
	basedir <- getBaseDirectory "state"
	best_effort_atomic_file_write path basedir filedata
-- 	log_debug_msg $ "wrote-map: " ++ (show filename) ++ " [" ++ (show $ length filedata) ++ "]"

get_state_tablename ref stateset = do
	let Just sessid = get_sessid $ cookie $ connection $ ref
	tablename <- do
		ai <- getApiInfo ref
		case stateset of
			"session" -> return $ printf "session_%d_%d_%s" (ascension ai) (daysthisrun ai) sessid
			"character" -> return "character"
			"ascension" -> return $ "ascension_" ++ (show $ ascension ai)
			"day" -> return $ "day_" ++ (show $ ascension ai) ++ "_" ++ (show $ daysthisrun ai)
			_ -> throwIO $ InternalError $ "Invalid state table type: " ++ stateset
-- 	doSTATE_DEBUG $ "tablename: " ++ tablename
	return tablename

registerUpdatedState ref stateset var = do
-- 	putStrLn $ "DEBUG: updstate " ++ stateset ++ "." ++ var
	case stateset of
		"character" -> void $ storeSettingsOnServer ref
		"ascension" -> void $ storeSettingsOnServer ref
		"day" -> case take 6 var of
			"fight." -> return ()
			_ -> void $ storeSettingsOnServer ref
		_ -> return ()
	let var = "turnsplayed last state change"
	ai <- getApiInfo ref
	let value = show $ turnsplayed ai
	tablename <- get_state_tablename ref "character"
	(doStateAction $ sessionData $ ref) ref $ \db -> do
		do_db_query_ db ("CREATE TABLE IF NOT EXISTS " ++ tablename ++ "(name TEXT NOT NULL, value TEXT NOT NULL, PRIMARY KEY(name));") []
		do_db_query_ db ("INSERT OR REPLACE INTO " ++ tablename ++ "(name, value) VALUES(?, ?);") [Just var, Just value]

setState ref stateset var value = do
	doSTATE_DEBUG $ "setState " ++ stateset ++ ":" ++ var ++ " => " ++ value
	Just (stid, st) <- readIORef $ state ref
	let newstate = case st of
		(requestmap, sessionmap, charmap, ascmap, daymap) ->
			case stateset of
				"request" -> (Data.Map.insert var value requestmap, sessionmap, charmap, ascmap, daymap)
				"session" -> (requestmap, Data.Map.insert var value sessionmap, charmap, ascmap, daymap)
				"character" -> (requestmap, sessionmap, Data.Map.insert var value charmap, ascmap, daymap)
				"ascension" -> (requestmap, sessionmap, charmap, Data.Map.insert var value ascmap, daymap)
				"day" -> (requestmap, sessionmap, charmap, ascmap, Data.Map.insert var value daymap)
		-- 		"session" -> 
		-- 		"fight" -> 
				_ -> throw $ InternalError $ "Invalid state table type: " ++ stateset
	writeIORef (state ref) $ Just (stid, newstate)
	tablename <- get_state_tablename ref stateset
	(doStateAction $ sessionData $ ref) ref $ \db -> do
		do_db_query_ db ("CREATE TABLE IF NOT EXISTS " ++ tablename ++ "(name TEXT NOT NULL, value TEXT NOT NULL, PRIMARY KEY(name));") []
		do_db_query_ db ("INSERT OR REPLACE INTO " ++ tablename ++ "(name, value) VALUES(?, ?);") [Just var, Just value]
	registerUpdatedState ref stateset var

unsetState ref stateset var = do
	doSTATE_DEBUG $ "unsetState " ++ stateset ++ ":" ++ var
	Just (stid, st) <- readIORef (state ref)
	let newstate = case st of
		(requestmap, sessionmap, charmap, ascmap, daymap) ->
			case stateset of
				"request" -> (Data.Map.delete var requestmap, sessionmap, charmap, ascmap, daymap)
				"session" -> (requestmap, Data.Map.delete var sessionmap, charmap, ascmap, daymap)
				"character" -> (requestmap, sessionmap, Data.Map.delete var charmap, ascmap, daymap)
				"ascension" -> (requestmap, sessionmap, charmap, Data.Map.delete var ascmap, daymap)
				"day" -> (requestmap, sessionmap, charmap, ascmap, Data.Map.delete var daymap)
		-- 		"fight" -> 
				_ -> throw $ InternalError $ "Invalid state table type: " ++ stateset
	writeIORef (state ref) $ Just (stid, newstate)
	tablename <- get_state_tablename ref stateset
	(doStateAction $ sessionData $ ref) ref $ \db -> do
		do_db_query_ db ("CREATE TABLE IF NOT EXISTS " ++ tablename ++ "(name TEXT NOT NULL, value TEXT NOT NULL, PRIMARY KEY(name));") []
		do_db_query_ db ("DELETE FROM " ++ tablename ++ " WHERE name = ?;") [Just var]
	registerUpdatedState ref stateset var

getState ref stateset var = do
	Just (_, st) <- readIORef (state ref)
	let value = case st of
		(requestmap, sessionmap, charmap, ascmap, daymap) ->
			case stateset of
				"request" -> Data.Map.lookup var requestmap
				"session" -> Data.Map.lookup var sessionmap
				"character" -> Data.Map.lookup var charmap
				"ascension" -> Data.Map.lookup var ascmap
				"day" -> Data.Map.lookup var daymap
		-- 		"fight" -> 
				_ -> throw $ InternalError $ "Invalid state table type: " ++ stateset
-- 	doSTATE_DEBUG $ "getState " ++ stateset ++ ":" ++ var ++ " <- " ++ show value
	return (value :: Maybe String)

uglyhack_resetFightState ref = do
	doSTATE_DEBUG $ "uglyhack_resetFightState"
-- 	print_debug_log_msg $ "== reset fight =="
	Just (_stid, (_requestmap, _sessionmap, _charmap, _ascmap, daymap)) <- readIORef (state ref)
	let fight_keys = Data.Map.toList $ Data.Map.filterWithKey (\x _y -> isPrefixOf "fight." x) daymap
	mapM_ (\(x, _y) -> unsetState ref "day" x) fight_keys

makeStateJSON ref = do
	doSTATE_DEBUG $ "makeStateJSON"
	ai <- getApiInfo ref
	let tnow = turnsplayed ai
	let charname = charName ai
	let ascnum = ascension ai
	let daynum = daysthisrun ai
	Just (_, (_requestmap, _sessionmap, charmap, ascmap, extra_daymap)) <- readIORef (state ref)
	let daymap = Data.Map.filterWithKey (\x _y -> not $ isPrefixOf "fight." x) extra_daymap
	let char = toJSObject [("name", JSString $ toJSString $ charname), ("state", JSString $ toJSString $ show charmap)]
	let asc = toJSObject [("ascension", JSString $ toJSString $ show ascnum), ("state", JSString $ toJSString $ show ascmap)]
	let day = toJSObject [("ascension-day", JSString $ toJSString $ (show ascnum) ++ "-" ++ (show daynum)), ("state", JSString $ toJSString $ show daymap)]
	let timestamp = toJSObject [("turnsplayed", JSString $ toJSString $ show $ tnow)]
-- 	putStrLn $ "new server settings timestamp: " ++ show (tnow, timestamp)
	return $ toJSObject [("character", JSObject char), ("ascension", JSObject asc), ("day", JSObject day), ("timestamp", JSObject timestamp)]

-- TODO: set a timestamp for local settings, overwrite if server is >= that, not just >= current!
loadStateJSON ref jsobj = do
	doSTATE_DEBUG $ "loadStateJSON"
	lstp <- getState ref "character" "turnsplayed last state change"
	ai <- getApiInfo ref
	let charname = charName ai
	let ascnum = ascension ai
	let daynum = daysthisrun ai
	doSTATE_DEBUG $ "loadStateJSON reading ioref"
	Just (stid, (requestmap, sessionmap, charmap, ascmap, daymap)) <- readIORef (state ref)
	doSTATE_DEBUG $ "loadStateJSON: " ++ show (stid)
	what_list <- do
		let Ok stuff = fromJSObject <$> readJSON jsobj

		case lookup "timestamp" stuff of
			Just tstampobj -> do
				let turnsplayedstr = fromJSString $ fromJust $ lookup "turnsplayed" $ fromJSObject tstampobj :: String
				let Just t = read_as $ turnsplayedstr :: Maybe Integer
				let should_use_server_data = case read_as =<< lstp :: Maybe Integer of
					Just tnow -> (t >= tnow)
					Nothing -> True
				if should_use_server_data
					then do
						putStrLn $ "server settings are recent " ++ show (t, lstp)
						putStrLn $ "loading character state"
						let charlist = fromJSObject $ fromJust $ lookup "character" stuff
						let storedcharname = fromJSString $ fromJust $ lookup "name" charlist
						(what, newstate) <- if storedcharname == charname
							then do
								let storedcharstatestr = fromJSString $ fromJust $ lookup "state" charlist
								let Just storedcharmap = read_as storedcharstatestr :: Maybe (Data.Map.Map String String)

								putStrLn $ "loading ascension state"
								let asclist = fromJSObject $ fromJust $ lookup "ascension" stuff
								let storedascnum = fromJSString $ fromJust $ lookup "ascension" asclist
								if (Just ascnum) == (read_as storedascnum :: Maybe Integer)
									then do
										let storedascstatestr = fromJSString $ fromJust $ lookup "state" asclist
										let Just storedascmap = read_as storedascstatestr :: Maybe (Data.Map.Map String String)

										putStrLn $ "loading day state"
										let daylist = fromJSObject $ fromJust $ lookup "day" stuff
										let storeddaynum = fromJSString $ fromJust $ lookup "ascension-day" daylist
										if storeddaynum == ((show ascnum) ++ "-" ++ (show daynum))
											then do
												let storeddaystatestr = fromJSString $ fromJust $ lookup "state" daylist
												let Just storeddaymap = read_as storeddaystatestr :: Maybe (Data.Map.Map String String)
												return (["character", "ascension", "day"], (requestmap, sessionmap, storedcharmap, storedascmap, storeddaymap))
											else return (["character", "ascension"], (requestmap, sessionmap, storedcharmap, storedascmap, daymap))
									else return (["character"], (requestmap, sessionmap, storedcharmap, ascmap, daymap))
							else return (["Nothing"], (requestmap, sessionmap, charmap, ascmap, daymap))
						doSTATE_DEBUG $ "loadStateJSON new: " ++ show (stid, newstate) ++ " # " ++ show what
						writeIORef (state ref) $ Just (stid, newstate)
						return what
					else do
						putStrLn $ "server settings are old " ++ show (t, lstp)
						return ["Nothing"]
			Nothing -> do
				putStrLn $ "no turnsplayed timestamp for server settings"
				return ["Nothing"]
	mirrorStateIntoDatabase ref
	return $ show what_list

mirrorStateIntoDatabase ref = do
	Just (_, (_requestmap, _sessionmap, charmap, ascmap, daymap)) <- readIORef (state ref)
	mv <- newEmptyMVar
	(doStateAction $ sessionData $ ref) ref $ \db -> do
		do_db_query_ db ("BEGIN") []
		let save_tbl base_tablename contents = do
			tablename <- get_state_tablename ref base_tablename
			do_db_query_ db ("CREATE TABLE IF NOT EXISTS " ++ tablename ++ "(name TEXT NOT NULL, value TEXT NOT NULL, PRIMARY KEY(name));") []
			do_db_query_ db ("DELETE FROM " ++ tablename ++ ";") []
			mapM_ (\(x, y) -> do
				do_db_query_ db ("INSERT OR REPLACE INTO " ++ tablename ++ "(name, value) VALUES(?, ?);") [Just x, Just y]
				return ()) $ Data.Map.toList contents
		save_tbl "character" charmap
		save_tbl "ascension" ascmap
		save_tbl "day" daymap
		do_db_query_ db ("COMMIT") []
		putStrLn $ "storing loaded settings."
		putMVar mv ()
	void $ takeMVar mv

storeSettingsOnServer ref = do
	Just (_, (_requestmap, _sessionmap, charmap, ascmap, extra_daymap)) <- readIORef (state ref)
	let daymap = Data.Map.filterWithKey (\x _y -> not $ isPrefixOf "fight." x) extra_daymap
	let statedesc = show (charmap, ascmap, daymap)
	laststored <- readIORef (lastStoredState $ sessionData $ ref)
	let shouldstore = case laststored of
		Just x -> x /= statedesc
		Nothing -> True
	when shouldstore $ do
		ai <- getApiInfo ref
		json <- Data.ByteString.Char8.unpack <$> getPageRawNoScripts ("/actionbar.php?action=fetch&for=kolproxy+" ++ kolproxy_version_number ++ "+by+Eleron&format=json") ref
		let Ok list = fromJSObject <$> decodeStrict json
		stateobj <- makeStateJSON ref
		let newlist = filter (\(x, _) -> x /= "kolproxy state") list ++ [("kolproxy state", JSObject stateobj)]
		let newjson = encodeStrict $ toJSObject $ newlist
		postPageRawNoScripts "/actionbar.php" [("action", "set"), ("for", "kolproxy " ++ kolproxy_version_number ++ " by Eleron"), ("format", "json"), ("bar", newjson), ("pwd", pwd ai)] ref
		putStrLn $ "stored settings on server (" ++ (show $ length newjson) ++ " bytes.)"
		writeIORef (lastStoredState $ sessionData $ ref) (Just statedesc)

loadSettingsFromServer ref = do
	json <- Data.ByteString.Char8.unpack <$> getPageRawNoScripts ("/actionbar.php?action=fetch&for=kolproxy+" ++ kolproxy_version_number ++ "+by+Eleron&format=json") ref
	case fromJSObject <$> decodeStrict json of
		Ok list -> case (lookup "kolproxy state" list) of
			Just x -> Just <$> loadStateJSON ref x
			Nothing -> return Nothing
		_ -> do
			putStrLn $ "Error loading actionbar data:"
			putStrLn $ ""
			putStrLn $ json
			putStrLn $ ""
			throwIO StateException
